"""
Data object for livetiming data
"""

import json
import warnings
from datetime import timedelta

from fastf1.logger import get_logger
from fastf1.utils import (
    recursive_dict_get,
    to_datetime
)


_logger = get_logger(__name__)


_track_status_mapping = {
    'AllClear': '1',
    'Yellow': '2',
    'SCDeployed': '4',
    'Red': '5',
    'VSCDeployed': '6',
    'VSCEnding': '7'
}


class LiveTimingData:
    """Live timing data object for using saved livetiming data as data source.

    This object is created from data that was recorded using
    :class:`~fastf1.livetiming.client.SignalRClient`. It can be passed to
    various api calling functions using the ``livedata`` keyword.

    Usually you will only instantiate this function and pass it to
    other functions.

    See :mod:`fastf1.livetiming` for a usage example.

    If you want to load data from multiple files you can simply pass multiple
    filenames::

        livedata = LiveTimingData('file1.txt', 'file2.txt', 'file3.txt')

    The files need to be in chronological order but may overlap. I.e. if the
    last five minutes of file1 are the same as the first 5 minutes of file2
    this will be recognized while loading the data. No duplicate data will
    be loaded.

    Args:
        *files (str): One or multiple file names
    """
    def __init__(self, *files, **kwargs):
        # file names
        self.files = files
        # parsed data
        self.data = dict()
        # number of json errors
        self.errorcount = 0
        # flag for automatic data loading on first access
        self._files_read = False
        # date when session was started
        self._start_date = None

        if 'remove_duplicates' in kwargs:
            warnings.warn("The argument `remove_duplicates` is no longer "
                          "available. Duplicates caused by overlapping files "
                          "will now always be removed.")

    def load(self):
        """
        Read all files, parse the data and store it by category.

        Should usually not be called manually. This is called
        automatically the first time :meth:`get`, :meth:`has`
        or :meth:`list_categories` are called.
        """
        _logger.info("Reading live timing data from recording. "
                     "This may take a bit.")

        is_first = True
        _files = [*self.files, None]
        current_data, next_data = None, None

        # We always need the current and next file loaded, so we can detect
        # where they overlap. The "next" file then becomes the "current" file
        # and the next "next" file is read.
        for next_file in _files:
            # make the previous "next" file the "current" file
            current_data = next_data

            if next_file is None:
                # reached the end, there is no subsequent data anymore
                next_data = None
            else:
                # read a new file as next file
                with open(next_file) as fobj:
                    next_data = fobj.readlines()

            if current_data is None:
                # there is no "current" file yet (i.e. first iteration),
                # skip ahead once right away to read one more file
                continue

            next_line = next_data[0] if next_data else None

            self._load_single_file(current_data,
                                   is_first_file=is_first,
                                   next_line=next_line)
            is_first = False

        # set flag that all files have been read
        self._files_read = True

    def _load_single_file(self, data, *, is_first_file, next_line):
        # parse its content and add it to the already loaded
        # data (if there is data already)

        # try to find the correct start date (only if this is the first file)
        if is_first_file:
            self._try_set_correct_start_date(data)

        for line in data:
            if line == next_line:
                break
            self._parse_line(line)

        # first file was loaded, others are appended if any more are loaded
        self._previous_files = True

    def _parse_line(self, elem):
        # parse a single line of data

        # load the three parts of each data element
        elem = self._fix_json(elem)
        try:
            cat, msg, dt_str = json.loads(elem)
        except (json.JSONDecodeError, ValueError):
            self.errorcount += 1
            return

        # convert string to datetime
        dt = to_datetime(dt_str)
        if dt is None:
            self.errorcount += 1
            return

        # if no start date could be determined beforehand, simply use the
        # first timestamp as we need to have some date as start date;
        # convert timestamp to timedelta (SessionTime) base on start date
        if self._start_date is None:
            self._start_date = dt
            td = timedelta(seconds=0)
        else:
            td = dt - self._start_date

        self._add_to_category(cat, [td, msg])

    def _fix_json(self, elem):
        # fix F1's not json compliant data
        elem = elem.replace("'", '"') \
            .replace('True', 'true') \
            .replace('False', 'false')
        return elem

    def _add_to_category(self, cat, entry):
        if cat not in self.data:
            self.data[cat] = [entry, ]
        else:
            self.data[cat].append(entry)

    def _try_set_correct_start_date(self, data):
        # skim content to find 'Started' session status without actually
        # decoding each line to save time
        for elem in data:
            if 'SessionStatus' in elem and 'Started' in elem:
                break
        else:
            # didn't find 'Started'
            _logger.error("Error while trying to set correct "
                          "session start date!")
            return

        # decode matching line
        elem = self._fix_json(elem)
        try:
            cat, msg, dt = json.loads(elem)
        except (json.JSONDecodeError, ValueError):
            _logger.error("Error while trying to set correct "
                          "session start date!")
            return

        # find correct entry in series
        try:
            for entry in msg['StatusSeries']:
                status = recursive_dict_get(entry, 'SessionStatus')
                if status == 'Started':
                    try:
                        self._start_date = to_datetime(entry['Utc'])
                    except (KeyError, ValueError, TypeError):
                        self.errorcount += 1
                        _logger.error("Error while trying to set correct "
                                      "session start date!")
                        return
        except AttributeError:
            for entry in msg['StatusSeries'].values():
                status = entry.get('SessionStatus', None)
                if status == 'Started':
                    try:
                        self._start_date = to_datetime(entry['Utc'])
                    except (KeyError, ValueError, TypeError):
                        self.errorcount += 1
                        _logger.error("Error while trying to set correct "
                                      "session start date!")
                        return

    def get(self, name):
        """
        Return data for category name.

        Will load data on first call, this will take a bit.

        Args:
            name (str): name of the category
            """
        if not self._files_read:
            self.load()
        return self.data[name]

    def has(self, name):
        """
        Check if data for a category name exists.

        Will load data on first call, this will take a bit.

        Args:
            name (str): name of the category
        """
        if not self._files_read:
            self.load()
        return name in self.data.keys()

    def list_categories(self):
        """
        List all available data categories.

        Will load data on first call, this will take a bit.

        Returns:
            list of category names
        """
        if not self._files_read:
            self.load()
        return list(self.data.keys())




from fastf1.plotting._constants.base import (
    CompoundsConst,
    TeamColorsConst,
    TeamConst
)


# NOTE: the team constants are copied when loading the driver-team-mapping
# and values may be modified there, it the used API provides different values


Teams: dict[str, TeamConst] = {
    'alfa romeo': TeamConst(
        ShortName='Alfa Romeo',
        TeamColor=TeamColorsConst(
            Official='#9b0000',
            FastF1='#900000'
        )
    ),
    'haas': TeamConst(
        ShortName='Haas',
        TeamColor=TeamColorsConst(
            Official='#bd9e57',
            FastF1='#bd9e57'
        )
    ),
    'ferrari': TeamConst(
        ShortName='Ferrari',
        TeamColor=TeamColorsConst(
            Official='#dc0000',
            FastF1='#da291c'
        )
    ),
    'mclaren': TeamConst(
        ShortName='McLaren',
        TeamColor=TeamColorsConst(
            Official='#ff8700',
            FastF1='#ff8000'
        )
    ),
    'mercedes': TeamConst(
        ShortName='Mercedes',
        TeamColor=TeamColorsConst(
            Official='#00d2be',
            FastF1='#00d2be'
        )
    ),
    'racing point': TeamConst(
        ShortName='Racing Point',
        TeamColor=TeamColorsConst(
            Official='#f596c8',
            FastF1='#ff87bc'
        )
    ),
    'red bull': TeamConst(
        ShortName='Red Bull',
        TeamColor=TeamColorsConst(
            Official='#1e41ff',
            FastF1='#1e41ff'
        )
    ),
    'renault': TeamConst(
        ShortName='Renault',
        TeamColor=TeamColorsConst(
            Official='#fff500',
            FastF1='#fff500'
        )
    ),
    'toro rosso': TeamConst(
        ShortName='Toro Rosso',
        TeamColor=TeamColorsConst(
            Official='#469bff',
            FastF1='#2b4562'
        )
    ),
    'williams': TeamConst(
        ShortName='Williams',
        TeamColor=TeamColorsConst(
            Official='#ffffff',
            FastF1='#00a0dd'
        )
    )
}

CompoundColors: dict[CompoundsConst, str] = {
    CompoundsConst.Soft: "#da291c",
    CompoundsConst.Medium: "#ffd12e",
    CompoundsConst.Hard: "#f0f0ec",
    CompoundsConst.Intermediate: "#43b02a",
    CompoundsConst.Wet: "#0067ad",
    CompoundsConst.Unknown: "#00ffff",
    CompoundsConst.TestUnknown: "#434649"
}



import warnings

import numpy as np


with warnings.catch_warnings():
    warnings.filterwarnings(
        'ignore', message="Using slow pure-python SequenceMatcher"
    )
    # suppress that warning, it's confusing at best here, we don't need fast
    # sequence matching and the installation (on windows) requires some effort
    from rapidfuzz import fuzz


def fuzzy_matcher(
        query: str,
        reference: list[list[str]],
        abs_confidence: float = 0.0,
        rel_confidence: float = 0.0
) -> (int, bool):
    """
    Match a query string to a reference list of lists of strings using fuzzy
    string matching.

    The reference is a list of sub-lists where each sub-list represents one
    element. The sub-lists contain one or multiple feature strings. The idea is
    that each element can be described by multiple feature strings. The
    function tries to find the best matching element in the reference list
    for the given query string.

    The function first checks for exact substring matches with the individual
    feature strings. If there is exactly one sub-list, where the query
    is a substring of a feature string, this index is returned as an
    "accurate match". Else, the function uses fuzzy string matching to find the
    best match in the reference list. The index of the best matching element is
    then returned as an "inaccurate match".

    Args:
        query: The query string to match.
        reference: A list of lists where each sub-list contains one or multiple
            feature strings describing an element.
        abs_confidence: The minimum absolute confidence that the match must
            have when fuzzy matched. Must be a value between 0.0 and 1.0, where
              1.0 is equivalent to a perfect match. Set to 0.0 to disable.
              If the best match has a lower confidence, a KeyError is raised.
        rel_confidence: The minimum relative confidence that the match must
            have (compared with the second-best match). Must be a value greater
            than 0.0, where 0.5 would mean that the best match must have a 50%
            higher score than the second-best match. Set to 0.0 to disable.
            If the best match has a lower relative confidence, a KeyError is
            raised.

    Returns:
        (int, bool): Index of the best matching element in the
            reference (outer) list and a boolean indicating if the match is
            accurate or not.

    """
    # Preprocess the query and reference strings
    query = query.casefold().replace(" ", "")
    for i in range(len(reference)):
        for j in range(len(reference[i])):
            reference[i][j] = reference[i][j].casefold().replace(" ", "")

    # Check for exact substring matches with the individual feature strings
    # first. If there is exactly one reference tuple, where the query is a
    # substring of a feature string, return this index as accurate match.
    full_partial_match_indices = []
    for i, feature_strings in enumerate(reference):
        if any([query in val for val in feature_strings]):
            full_partial_match_indices.append(i)

    if len(full_partial_match_indices) == 1:
        # return index as accurate match
        return full_partial_match_indices[0], True

    # Zero or multiple reference tuples had substring matches, so we need to
    # do fuzzy matching
    reference = np.array(reference)
    ratios = np.zeros_like(reference, dtype=int)

    # If we have multiple substring matches, we only fuzzy match on these,
    # else we fuzzy match on all reference tuples
    if full_partial_match_indices:
        candidate_indices = full_partial_match_indices
    else:
        candidate_indices = range(len(reference))

    # Calculate the fuzz ratio for each feature string in each reference tuple
    for i in candidate_indices:
        feature_strings = reference[i]
        ratios[i] = [fuzz.ratio(val, query) for val in feature_strings]

    max_ratio = np.max(ratios)
    max_row_ratios = np.max(ratios, axis=1)
    # if there are multiple rows with the same maximum ratio, we need to remove
    # the corresponding ratios from the comparison so that we can match based
    # on the remaining feature string ratios
    if np.sum(max_row_ratios == max_ratio) > 1:
        # get counts of all unique ratios and remove all that are not unique
        # in the array by setting them to zero
        unique, counts = np.unique(reference, return_counts=True)
        count_dict = dict(zip(unique, counts))
        mask = ((np.vectorize(count_dict.get)(reference) > 1)
                & (ratios == max_ratio))
        ratios[mask] = 0

    # get the index of the row that contains the maximum ratio
    max_index = np.argmax(ratios) // ratios.shape[1]

    # optional confidence checks
    if abs_confidence and (max_ratio < (abs_confidence * 100)):
        raise KeyError(f"Found no match for '{query}' with sufficient "
                       f"absolute confidence")

    if rel_confidence and (max_ratio / np.partition(ratios.flatten(), -2)[-2]
                           < (1 + rel_confidence)):
        # max ratio divided by second-largest ratio is less
        # than 1 + rel_confidence
        raise KeyError(f"Found no match for '{query}' with sufficient "
                       f"relative confidence")

    # return index as inaccurate match
    return max_index, False



from dataclasses import dataclass


class CompoundsConst:
    HyperSoft = "HYPERSOFT"
    UltraSoft = "ULTRASOFT"
    SuperSoft = "SUPERSOFT"
    Soft = "SOFT"
    Medium = "MEDIUM"
    Hard = "HARD"
    SuperHard = "SUPERHARD"
    Intermediate = "INTERMEDIATE"
    Wet = "WET"
    Unknown = "UNKNOWN"
    TestUnknown = "TEST-UNKNOWN"


@dataclass(frozen=True)
class TeamColorsConst:
    Official: str
    FastF1: str


@dataclass(frozen=True)
class TeamConst:
    ShortName: str
    TeamColor: TeamColorsConst


@dataclass(frozen=True)
class BaseSeasonConst:
    CompoundColors: dict[str, str]
    Teams: dict[str, TeamConst]



import unicodedata

from fastf1.logger import get_logger
from fastf1.plotting._constants.base import TeamConst


_logger = get_logger(__package__)


class _Driver:
    value: str = ''
    normalized_value: str = ''
    abbreviation: str = ''
    team: "_Team"


class _Team:
    value: str = ''
    normalized_value: str = ''
    constants: TeamConst = None

    def __init__(self):
        super().__init__()
        self.drivers: list[_Driver] = list()


class _DriverTeamMapping:
    def __init__(
            self,
            year: str,
            teams: list[_Team],
    ):
        self.year = year
        self.teams = teams

        self.drivers_by_normalized: dict[str, _Driver] = dict()
        self.drivers_by_abbreviation: dict[str, _Driver] = dict()
        self.teams_by_normalized: dict[str, _Team] = dict()

        for team in teams:
            for driver in team.drivers:
                self.drivers_by_normalized[driver.normalized_value] = driver
                self.drivers_by_abbreviation[driver.abbreviation] = driver
            self.teams_by_normalized[team.normalized_value] = team


def _normalize_string(name: str) -> str:
    # removes accents from a string and returns the closest possible
    # ascii representation (https://stackoverflow.com/a/518232)
    stripped = ''.join(c for c in unicodedata.normalize('NFD', name)
                       if unicodedata.category(c) != 'Mn')
    return stripped




from fastf1.plotting._constants.base import (
    CompoundsConst,
    TeamColorsConst,
    TeamConst
)


# NOTE: the team constants are copied when loading the driver-team-mapping
# and values may be modified there, if the used API provides different values


Teams: dict[str, TeamConst] = {
    'alpine': TeamConst(
        ShortName='Alpine',
        TeamColor=TeamColorsConst(
            Official='#0093cc',
            FastF1='#ff87bc'
        )
    ),
    'aston martin': TeamConst(
        ShortName='Aston Martin',
        TeamColor=TeamColorsConst(
            Official='#229971',
            FastF1='#00665f'
        )
    ),
    'ferrari': TeamConst(
        ShortName='Ferrari',
        TeamColor=TeamColorsConst(
            Official='#e80020',
            FastF1='#e80020'
        )
    ),
    'haas': TeamConst(
        ShortName='Haas',
        TeamColor=TeamColorsConst(
            Official='#b6babd',
            FastF1='#b6babd'
        )
    ),
    'mclaren': TeamConst(
        ShortName='McLaren',
        TeamColor=TeamColorsConst(
            Official='#ff8000',
            FastF1='#ff8000'
        )
    ),
    'mercedes': TeamConst(
        ShortName='Mercedes',
        TeamColor=TeamColorsConst(
            Official='#27f4d2',
            FastF1='#27f4d2'
        )
    ),
    'racing bulls': TeamConst(
        ShortName='RB',
        TeamColor=TeamColorsConst(
            Official='#6692ff',
            FastF1='#fcd700'
        )
    ),
    'red bull': TeamConst(
        ShortName='Red Bull',
        TeamColor=TeamColorsConst(
            Official='#3671c6',
            FastF1='#0600ef'
        )
    ),
    'kick sauber': TeamConst(
        ShortName='Sauber',
        TeamColor=TeamColorsConst(
            Official='#52e252',
            FastF1='#00e700'
        )
    ),
    'williams': TeamConst(
        ShortName='Williams',
        TeamColor=TeamColorsConst(
            Official='#64c4ff',
            FastF1='#00a0dd'
        )
    )
}

CompoundColors: dict[CompoundsConst, str] = {
    CompoundsConst.Soft: "#da291c",
    CompoundsConst.Medium: "#ffd12e",
    CompoundsConst.Hard: "#f0f0ec",
    CompoundsConst.Intermediate: "#43b02a",
    CompoundsConst.Wet: "#0067ad",
    CompoundsConst.Unknown: "#00ffff",
    CompoundsConst.TestUnknown: "#434649"
}
